----------------------------------------------------------------------------
-- Euclidean modular divider (Euclidean_divider.vhd)
--
-- Instantiate 3 components:
--          - non-restoring divider
--          - mult_subt
--          - non-restoring Reducer
----------------------------------------------------------------------------

library IEEE;
use IEEE.std_logic_1164.all;
use IEEE.std_logic_arith.all;
use IEEE.std_logic_unsigned.all;
package Euclidean_divider_parameters is
  constant K: natural := 8;
  constant P: std_logic_vector(K-1 downto 0) := conv_std_logic_vector(239, K);
  constant COUNTER_SIZE: natural := 3;
--  constant K: natural := 192;
--  constant P: std_logic_vector(K-1 downto 0) := x"fffffffffffffffffffffffffffffffeffffffffffffffff";
--  constant COUNTER_SIZE: natural := 8;
--  constant K: natural := 256;
--  constant P: std_logic_vector(K-1 downto 0) := (223 downto 193 => '0',191 downto 96 => '0',others =>'1');

  constant N: natural := 2*K-1;
  constant CS_ZERO: std_logic_vector(COUNTER_SIZE-1 downto 0) := (others => '0');
  constant ZERO: std_logic_vector(k-1 downto 0) := (others => '0');
  constant SHORT_ZERO: std_logic_vector(K-2 downto 0) := (others => '0');
  constant LONG_ZERO: std_logic_vector(2*K-1 downto 0) := (others => '0');
  constant one: std_logic_vector(K-1 downto 0) := conv_std_logic_vector(1, K);
end Euclidean_divider_parameters;

----------------------------------------------------------------------------
--  non-restoring divider
----------------------------------------------------------------------------
library IEEE;
use IEEE.std_logic_1164.all;
use IEEE.std_logic_arith.all;
use IEEE.std_logic_unsigned.all;
use work.Euclidean_divider_parameters.all;
entity nr_divider is
port (
  a, b: in std_logic_vector (K-1 downto 0);
  clk, reset, start: in std_logic; 
  quotient, remainder: out std_logic_vector (K-1 downto 0);
  done: out std_logic
);
end nr_divider;

architecture rtl of nr_divider is
  signal s: std_logic_vector(2*k-1 downto 0);
  signal r: std_logic_vector(2*K-2 downto 0);
  signal q: std_logic_vector(k-1 downto 1);
  signal modified_q: std_logic_vector(k-1 downto 0);

  signal load, update, equal_ZERO: std_logic;
  signal count: std_logic_vector(COUNTER_SIZE -1 downto 0);
  type states is range 0 to 4;
  signal current_state: states;

begin

  parallel_register: process(clk)
  begin
  if clk'event and clk = '1' then
    if load = '1' then 
      s <= ZERO & a;
    elsif update = '1' then 
      s(2*K-1 downto 1) <= r(2*K-2 downto 0);
      s(0) <= '0';
    end if;
  end if;
  end process parallel_register;

  shift_register: process(clk)
  begin
  if clk'event and clk = '1' then
    if load = '1' then 
      q <= SHORT_ZERO; 
    elsif update = '1' then 
      for i in k-1 downto 2 loop q(i) <= q(i-1); end loop;
      q(1) <= not(s(2*K-1));
    end if;
  end if;
  end process shift_register;

  with s(2*K-1) select r(2*K-2 downto k-2) <= s(2*K-2 downto K-2) + b when '1', s(2*K-2 downto K-2) - b when others;
  r(K-3 downto 0) <= s(K-3 downto 0);

  modified_q <= not(q(K-1))&q(K-2 downto 1)&'1';
  with s(2*K-1) select quotient <=  modified_q when '0', modified_q -1 when others;
  with s(2*K-1) select remainder <= s(2*K-2 downto K-1) when '0', s(2*K-2 downto K-1) + b when others;

  counter: process(clk)
  begin
  if clk'event and clk = '1' then
    if load = '1' then 
      count <= conv_std_logic_vector(K-2, COUNTER_SIZE);
    elsif update = '1' then 
      count <= count-1; 
    end if;
  end if;
  end process counter;

  with count select equal_ZERO <= '1' when CS_ZERO, '0' when others;

  control_unit: process(clk, reset, current_state, equal_ZERO)
  begin
  case current_state is
    when 0 to 1 => load <= '0'; update <= '0'; done <= '1';
    when 2 => load <= '1'; update <= '0'; done <= '0';
    when 3 => load <= '0'; update <= '1'; done <= '0';
    when 4 => load <= '0'; update <= '0'; done <= '0';
  end case;

  if reset = '1' then 
    current_state <= 0;
  elsif clk'event and clk = '1' then
    case current_state is
      when 0 => if start = '0' then current_state <= 1; end if;
      when 1 => if start = '1' then current_state <= 2; end if;
      when 2 => current_state <= 3;
      when 3 => if equal_ZERO = '1' then current_state <= 4; end if;
      when 4 => current_state <= 0;
    end case;
  end if;
  end process;
end rtl; --nr_divider

----------------------------------------------------------------------------
--  mult_subt
----------------------------------------------------------------------------
library IEEE;
use IEEE.std_logic_1164.all;
use IEEE.std_logic_arith.all;
use IEEE.std_logic_unsigned.all;
use work.Euclidean_divider_parameters.all;
entity mult_subt is
port (
  c, d: in std_logic_vector (2*K-1 downto 0);
  q: in std_logic_vector(K-1 downto 0);
  clk, reset, start: in std_logic; 
  z: out std_logic_vector (2*K-1 downto 0);
  done: out std_logic
);
end mult_subt;

architecture rtl of mult_subt is
  signal u: std_logic_vector(2*K-1 downto 0);
  signal v: std_logic_vector(K-1 downto 0);
  signal dif: std_logic_vector(2*K downto 0);
  signal load, update, equal_ZERO: std_logic;
  signal count: std_logic_vector(COUNTER_SIZE -1 downto 0);
  type states is range 0 to 3;
  signal current_state: states;
begin

  parallel_register: process(clk)
  begin
  if clk'event and clk = '1' then
    if load = '1' then 
      u <= c;
    elsif update = '1' then 
      u <= dif(2*K downto 1);
    end if;
  end if;
  end process parallel_register;
  
  shift_register: process(clk)
  begin
  if clk'event and clk = '1' then
    if load = '1' then 
      v <= q; 
    elsif update = '1' then 
      for i in 0 to K-2 loop v(i) <= v(i+1); end loop;
      v(K-1) <= dif(0);
    end if;
  end if;
  end process shift_register;
  
  with v(0) select dif <= (u(2*K-1)&u) - (d(2*K-1)&d) when '1', 
                                         (u(2*K-1)&u) when others;
  z <= u(K-1 downto 0) & v;

  counter: process(clk)
  begin
  if clk'event and clk = '1' then
    if load = '1' then 
      count <= conv_std_logic_vector(K-1, COUNTER_SIZE);
    elsif update = '1' then 
      count <= count-1; 
    end if;
  end if;
  end process counter;

  with count select equal_ZERO <= '1' when CS_ZERO, '0' when others;

  control_unit: process(clk, reset, current_state, equal_ZERO)
  begin
  case current_state is
    when 0 to 1 => load <= '0'; update <= '0'; done <= '1';
    when 2 => load <= '1'; update <= '0'; done <= '0';
    when 3 => load <= '0'; update <= '1'; done <= '0';
  end case;
  if reset = '1' then current_state <= 0;
  elsif clk'event and clk = '1' then
    case current_state is
      when 0 => if start = '0' then current_state <= 1; end if;
      when 1 => if start = '1' then current_state <= 2; end if;
      when 2 => current_state <= 3;
      when 3 => if equal_ZERO = '1' then current_state <= 0; end if;
    end case;
  end if;
  end process;
end rtl;

----------------------------------------------------------------------------
--  non-restoring Reducer
----------------------------------------------------------------------------
library IEEE;
use IEEE.std_logic_1164.all;
use IEEE.std_logic_arith.all;
use IEEE.std_logic_unsigned.all;
use work.Euclidean_divider_parameters.all;
entity nr_reducer is
port (
  x: in std_logic_vector (N downto 0);
  m: in std_logic_vector(K-1 downto 0);
  clk, reset, start: in std_logic; 
  z: out std_logic_vector (K-1 downto 0);
  done: out std_logic
);
end nr_reducer;

architecture rtl of nr_reducer is
  signal s: std_logic_vector(n+1 downto 0);
  signal r: std_logic_vector(n downto 0);
  signal minus_m, w: std_logic_vector(k downto 0);
  signal not_m: std_logic_vector(K-1 downto 0);
  signal load, update, equal_ZERO: std_logic;
  signal count: std_logic_vector(COUNTER_SIZE -1 downto 0);
  type states is range 0 to 3;
  signal current_state: states;

begin

  r(n downto N-K) <= s(n downto N-K) + w;
  r(N-K-1 downto 0) <= s(N-K-1 downto 0) ;
  with r(n) select z <= r(N-1 downto N-K) when '0', r(N-1 downto N-K) + m when others;

  registers: process(clk)
  begin
  if clk'event and clk = '1' then
    if load = '1' then 
      s <= x(N) & x; 
    elsif update = '1' then 
      s(0) <= '0';
      for i in 1 to N+1 loop s(i) <= r(i-1); end loop;
    end if;
  end if;
  end process registers;
  
  counter: process(clk)
  begin
  if clk'event and clk = '1' then
    if load = '1' then 
      count <= conv_std_logic_vector(N-K-1, COUNTER_SIZE);
    elsif update = '1' then 
      count <= count-1; 
    end if;
  end if;
  end process counter;

  with count select equal_ZERO <= '1' when CS_ZERO, '0' when others;
  not_gates: for i in 0 to K-1 generate not_m(i) <= not(m(i)); end generate;
  minus_m <= ('1'&not_m) +1;
  with s(n+1) select w <= minus_m when '0', ('0'&m) when others;
  
  control_unit: process(clk, reset, current_state, equal_ZERO)
  begin
  case current_state is
    when 0 to 1 => load <= '0'; update <= '0'; done <= '1';
    when 2 => load <= '1'; update <= '0'; done <= '0';
    when 3 => load <= '0'; update <= '1'; done <= '0';
  end case;
  if reset = '1' then 
    current_state <= 0;
  elsif clk'event and clk = '1' then
    case current_state is
      when 0 => if start = '0' then current_state <= 1; end if;
      when 1 => if start = '1' then current_state <= 2; end if;
      when 2 => current_state <= 3;
      when 3 => if equal_ZERO = '1' then current_state <= 0; end if;
    end case;
  end if;
  end process;
end rtl; --nr_reducer



----------------------------------------------------------------------------
--  Euclidean Divider
----------------------------------------------------------------------------
library IEEE;
use IEEE.std_logic_1164.all;
use IEEE.std_logic_arith.all;
use IEEE.std_logic_unsigned.all;
use work.Euclidean_divider_parameters.all;
entity Euclidean_divider is
port(
  x, y: in std_logic_vector(K-1 downto 0);
  clk, reset, start: in std_logic;  
  z: out std_logic_vector(K-1 downto 0);
  done: out std_logic
);
end Euclidean_divider;

architecture circuit of Euclidean_divider is

  component nr_divider is
  port (
    a, b: in std_logic_vector (K-1 downto 0);
    clk, reset, start: in std_logic; 
    quotient, remainder: out std_logic_vector (K-1 downto 0);
    done: out std_logic
  );
  end component;

  component mult_subt is
  port (
    c, d: in std_logic_vector (2*K-1 downto 0);
    q: in std_logic_vector(K-1 downto 0);
    clk, reset, start: in std_logic; 
    z: out std_logic_vector (2*K-1 downto 0);
    done: out std_logic
  );
  end component;

  component nr_reducer is
  port (
    x: in std_logic_vector (n downto 0);
    m: in std_logic_vector(K-1 downto 0);
    clk, reset, start: in std_logic; 
    z: out std_logic_vector (K-1 downto 0);
    done: out std_logic
  );
  end component;

  signal a, b, q, r: std_logic_vector(K-1 downto 0);
  signal c, d, next_d: std_logic_vector(2*K-1 downto 0);

  signal load, update, start_division, start_product, start_reduction,
  division_done, product_done, reduction_done, b_equal_1: std_logic;

  type states is range 0 to 10;
  signal current_state: states;

begin

  divider: nr_divider port map(a, b, clk, reset, start_division, q, r, division_done);
  multiplier: mult_subt port map(c, d, q, clk, reset, start_product, next_d, product_done);
  reducer: nr_reducer port map(d, p, clk, reset, start_reduction, z, reduction_done);

  registers: process(clk)
  begin
  if clk'event and clk = '1' then
    if load = '1' then 
      a <= p; b <= y; c <= LONG_ZERO; d <= ZERO&x;
    elsif update = '1' then
      a <= b; b <= r; c <= d; d <= next_d;
    end if;
  end if;
  end process registers;

  b_equal_1 <= '1' when b = one else '0';

  control_unit: process(clk, reset, current_state, b_equal_1)
  begin
  case current_state is
    when 0 to 1 => load <= '0'; update <= '0'; start_division <= '0'; start_product <= '0';
    start_reduction <= '0'; done <= '1';
    when 2 => load <= '1'; update <= '0'; start_division <= '0'; start_product <= '0';
    start_reduction <= '0'; done <= '0';

    when 3 => load <= '0'; update <= '0'; start_division <= '1'; start_product <= '0';
    start_reduction <= '0'; done <= '0';
    when 4 => load <= '0'; update <= '0'; start_division <= '0'; start_product <= '0';
    start_reduction <= '0'; done <= '0';

    when 5 => load <= '0'; update <= '0'; start_division <= '0'; start_product <= '1';
    start_reduction <= '0'; done <= '0';
    when 6 => load <= '0'; update <= '0'; start_division <= '0'; start_product <= '0';
    start_reduction <= '0'; done <= '0';
    when 7 => load <= '0'; update <= '1'; start_division <= '0'; start_product <= '0';
    start_reduction <= '0'; done <= '0';

    when 8 => load <= '0'; update <= '0'; start_division <= '0'; start_product <= '0';
    start_reduction <= '0'; done <= '0';

    when 9 => load <= '0'; update <= '0'; start_division <= '0'; start_product <= '0';
    start_reduction <= '1'; done <= '0';
    when 10 => load <= '0'; update <= '0'; start_division <= '0'; start_product <= '0';
    start_reduction <= '0'; done <= '0';
  end case;

  if reset = '1' then current_state <= 0;
  elsif clk'event and clk = '1' then
    case current_state is
      when 0 => if start = '0' then current_state <= 1; end if;
      when 1 => if start = '1' then current_state <= 2; end if;
      when 2 => current_state <= 3;

      when 3 => current_state <= 4;
      when 4 => if division_done = '1' then current_state <= 5; end if;

      when 5 => current_state <= 6;
      when 6 => if product_done = '1' then current_state <= 7; end if;
      when 7 => current_state <= 8;

      when 8 => if b_equal_1 = '0' then current_state <= 3; else current_state <= 9; end if;

      when 9 => current_state <= 10;
      when 10 => if reduction_done = '1' then current_state <= 0; end if;
    end case;
  end if;
  end process;


end circuit;


 