----------------------------------------------------------------------------
-- multitplication and subtraction (mult_subt.vhd)
--
-- Used in chapter 4 euclidean modular divider
--
--
----------------------------------------------------------------------------
library IEEE;
use IEEE.std_logic_1164.all;
use IEEE.std_logic_arith.all;
use IEEE.std_logic_unsigned.all;
package mult_subt_parameters is
  constant K: natural := 8;
  constant COUNTER_SIZE: natural := 3;
  constant CS_ZERO: std_logic_vector(COUNTER_SIZE-1 downto 0) := (others => '0');
end mult_subt_parameters;

----------------------------------------------------------------------------
--  mult_subt
----------------------------------------------------------------------------
library IEEE;
use IEEE.std_logic_1164.all;
use IEEE.std_logic_arith.all;
use IEEE.std_logic_unsigned.all;
use work.mult_subt_parameters.all;
entity mult_subt is
port (
  c, d: in std_logic_vector (2*K-1 downto 0);
  q: in std_logic_vector(K-1 downto 0);
  clk, reset, start: in std_logic; 
  z: out std_logic_vector (2*K-1 downto 0);
  done: out std_logic
);
end mult_subt;

architecture rtl of mult_subt is
  signal u: std_logic_vector(2*K-1 downto 0);
  signal v: std_logic_vector(K-1 downto 0);
  signal dif: std_logic_vector(2*K downto 0);

  signal load, update, equal_zero: std_logic;
  signal count: std_logic_vector(COUNTER_SIZE -1 downto 0);
  type states is range 0 to 3;
  signal current_state: states;

begin

  parallel_register: process(clk)
  begin
  if clk'event and clk = '1' then
    if load = '1' then 
      u <= c;
    elsif update = '1' then 
      u <= dif(2*k downto 1);
    end if;
  end if;
  end process parallel_register;

  shift_register: process(clk)
  begin
  if clk'event and clk = '1' then
    if load = '1' then 
      v <= q; 
    elsif update = '1' then 
      for i in 0 to k-2 loop v(i) <= v(i+1); end loop;
      v(k-1) <= dif(0);
    end if;
  end if;
  end process shift_register;

  with v(0) select dif <= (u(2*K-1) & u) - (d(2*K-1) & d) when '1', 
                          (u(2*k-1)&u) when others;
  z <= u(K-1 downto 0)&v;

  counter: process(clk)
  begin
  if clk'event and clk = '1' then
    if load = '1' then 
      count <= conv_std_logic_vector(K-1, COUNTER_SIZE);
    elsif update = '1' then 
      count <= count-1; 
    end if;
  end if;
  end process counter;
  with count select equal_zero <= '1' when CS_ZERO, '0' when others;

  control_unit: process(clk, reset, current_state, equal_zero)
  begin
  case current_state is
    when 0 to 1 => load <= '0'; update <= '0'; done <= '1';
    when 2 => load <= '1'; update <= '0'; done <= '0';
    when 3 => load <= '0'; update <= '1'; done <= '0';
  end case;

  if reset = '1' then 
    current_state <= 0;
  elsif clk'event and clk = '1' then
    case current_state is
      when 0 => if start = '0' then current_state <= 1; end if;
      when 1 => if start = '1' then current_state <= 2; end if;
      when 2 => current_state <= 3;
      when 3 => if equal_zero = '1' then current_state <= 0; end if;
    end case;
  end if;
  end process;

end rtl;

